@@ -4,12 +4,12 @@
 #
 # Copyright © 2009 Mikkel Kamstrup Erlandsen <mikkel.kamstrup@gmail.com>
 # Copyright © 2009-2010 Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
-# Copyright © 2009-2010 Seif Lotfy <seif@lotfy.com>
+# Copyright © 2009-2011 Seif Lotfy <seif@lotfy.com>
 # Copyright © 2009 Markus Korn <thekorn@gmx.net>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
+# the Free Software Foundation, either version 2.1 of the License, or
 # (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
@@ -24,11 +24,7 @@ import sqlite3
 import time
 import sys
 import os
-import math
-import gettext
 import logging
-import operator
-from itertools import islice
 from collections import defaultdict
 
 from zeitgeist.datamodel import Event as OrigEvent, StorageState, TimeRange, \
@@ -38,12 +34,13 @@ from _zeitgeist.engine.extension import ExtensionsCollection, get_extensions
 from _zeitgeist.engine import constants
 from _zeitgeist.engine.sql import get_default_cursor, unset_cursor, \
 	TableLookup, WhereClause
-	
-WINDOW_SIZE = 7
+from _zeitgeist.lrucache import LRUCache
 
-logging.basicConfig(level=logging.DEBUG)
 log = logging.getLogger("zeitgeist.engine")
 
+# we only allow replacement of half of the cache at once.
+MAX_CACHE_BATCH_SIZE = constants.CACHE_SIZE/2
+
 class NegationNotSupported(ValueError):
 	pass
 
@@ -125,6 +122,7 @@ class ZeitgeistEngine:
 		self._manifestation = TableLookup(cursor, "manifestation")
 		self._mimetype = TableLookup(cursor, "mimetype")
 		self._actor = TableLookup(cursor, "actor")
+		self._event_cache = LRUCache(constants.CACHE_SIZE)
 	
 	@property
 	def extensions(self):
@@ -149,60 +147,138 @@ class ZeitgeistEngine:
 		event[0][Event.Id] = row["id"] # Id property is read-only in the public API
 		event.timestamp = row["timestamp"]
 		for field in ("interpretation", "manifestation", "actor"):
-			setattr(event, field, getattr(self, "_" + field).value(row[field]))
+			# Try to get event attributes from row using the attributed field id
+			# If attribute does not exist we break the attribute fetching and return
+			# None instead of of crashing
+			try:
+				setattr(event, field, getattr(self, "_" + field).value(row[field]))
+			except KeyError, e:
+				log.error("Event %i broken: Table %s has no id %i" \
+						%(row["id"], field, row[field]))
+				return None
 		event.payload = row["payload"] or "" # default payload: empty string
 		return event
 	
 	def _get_subject_from_row(self, row):
 		subject = Subject()
-		for field in ("uri", "origin", "text", "storage"):
+		for field in ("uri", "text", "storage"):
 			setattr(subject, field, row["subj_" + field])
+		setattr(subject, "origin", row["subj_origin_uri"])
 		for field in ("interpretation", "manifestation", "mimetype"):
-			setattr(subject, field,
-				getattr(self, "_" + field).value(row["subj_" + field]))
+			# Try to get subject attributes from row using the attributed field id
+			# If attribute does not exist we break the attribute fetching and return
+			# None instead of crashing
+			try:
+				setattr(subject, field,
+					getattr(self, "_" + field).value(row["subj_" + field]))
+			except KeyError, e:
+				log.error("Event %i broken: Table %s has no id %i" \
+						%(row["id"], field, row["subj_" + field]))
+				return None
 		return subject
 	
-	def get_events(self, ids=None, rows=None, sender=None):
+	def get_events(self, ids, sender=None):
 		"""
 		Look up a list of events.
 		"""
 		
 		t = time.time()
 		
-		if not ids and not rows:
+		if not ids:
 			return []
 		
-		if ids:
-			rows = self._cursor.execute("""
-				SELECT * FROM event_view
-				WHERE id IN (%s)
-				""" % ",".join("%d" % id for id in ids)).fetchall()
-		else:
-			ids = (row[0] for row in rows)
+		# Split ids into cached and uncached
+		uncached_ids = []
+		cached_ids = []
 		
+		# If ids batch greater than MAX_CACHE_BATCH_SIZE ids ignore cache
+		use_cache = True
+		if len(ids) > MAX_CACHE_BATCH_SIZE:
+			use_cache = False
+		if not use_cache:
+			uncached_ids = ids
+		else:
+			for id in ids:
+				if id in self._event_cache:
+					cached_ids.append(id)
+				else:
+					uncached_ids.append(id)
+		
+		id_hash = defaultdict(list)
+		for n, id in enumerate(ids):
+			# the same id can be at multible places (LP: #673916)
+			# cache all of them
+			id_hash[id].append(n)
+		
+		# If we are not able to get an event by the given id
+		# append None instead of raising an Error. The client
+		# might simply have requested an event that has been
+		# deleted
 		events = {}
+		sorted_events = [None]*len(ids)
+		
+		for id in cached_ids:
+			event = self._event_cache[id]
+			if event:
+				event = self.extensions.apply_get_hooks(event, sender)
+				if event is not None:
+					for n in id_hash[event.id]:
+						# insert the event into all necessary spots (LP: #673916)
+						sorted_events[n] = event
+		
+		# Get uncached events
+		rows = self._cursor.execute("""
+			SELECT * FROM event_view
+			WHERE id IN (%s)
+			""" % ",".join("%d" % id for id in uncached_ids)).fetchall()
+		
+		log.debug("Got %d raw events in %fs" % (len(rows), time.time()-t))
+		t = time.time()
+		
+		t_get_event = 0
+		t_get_subject = 0
+		t_apply_get_hooks = 0
+		
 		for row in rows:
 			# Assumption: all rows of a same event for its different
 			# subjects are in consecutive order.
+			t_get_event -= time.time()
 			event = self._get_event_from_row(row)
-			if event.id not in events:
-				events[event.id] = event
-			events[event.id].append_subject(self._get_subject_from_row(row))
-		
-		# Sort events into the requested order
-		sorted_events = []
-		for id in ids:
-			# if we are not able to get an event by the given id
-			# append None instead of raising an Error. The client
-			# might simply have requested an event that has been
-			# deleted
-			event = events.get(id, None)
-			event = self.extensions.apply_get_hooks(event, sender)
+			t_get_event += time.time()
 			
-			sorted_events.append(event)
-		
-		log.debug("Got %d events in %fs" % (len(sorted_events), time.time()-t))
+			if event:
+				# Check for existing event.id in event to attach 
+				# other subjects to it
+				if event.id not in events:
+					events[event.id] = event
+				else:
+					event = events[event.id]
+					
+				t_get_subject -= time.time()
+				subject = self._get_subject_from_row(row)
+				t_get_subject += time.time()
+				# Check if subject has a proper value. If none than something went
+				# wrong while trying to fetch the subject from the row. So instead
+				# of failing and raising an error. We silently skip the event.
+				if subject:
+					event.append_subject(subject)
+					if use_cache and not event.payload:
+						self._event_cache[event.id] = event
+					t_apply_get_hooks -= time.time()
+					event = self.extensions.apply_get_hooks(event, sender)
+					t_apply_get_hooks += time.time()
+					if event is not None:
+						for n in id_hash[event.id]:
+							# insert the event into all necessary spots (LP: #673916)
+							sorted_events[n] = event
+					# Avoid caching events with payloads to have keep the cache MB size 
+					# at a decent level
+					
 
+		log.debug("Got %d events in %fs" % (len(sorted_events), time.time()-t))
+		log.debug("    Where time spent in _get_event_from_row in %fs" % (t_get_event))
+		log.debug("    Where time spent in _get_subject_from_row in %fs" % (t_get_subject))
+		log.debug("    Where time spent in apply_get_hooks in %fs" % (t_apply_get_hooks))
 		return sorted_events
 	
 	@staticmethod
@@ -306,9 +382,17 @@ class ZeitgeistEngine:
 		if storage_state != StorageState.Any:
 			raise NotImplementedError
 		
+		# thekorn: we are using the unary operator here to tell sql to not use
+		# the index on the timestamp column at the first place. This `fix` for
+		# (LP: #672965) is based on some benchmarks, which suggest a performance
+		# win, but we might not oversee all implications.
+		# (see http://www.sqlite.org/optoverview.html section 6.0)
 		where = WhereClause(WhereClause.AND)
-		where.add("timestamp >= ?", time_range[0])
-		where.add("timestamp <= ?", time_range[1])
+		min_time, max_time = time_range
+		if min_time != 0:
+			where.add("+timestamp >= ?", min_time)
+		if max_time != sys.maxint:
+			where.add("+timestamp <= ?", max_time)
 		
 		where.extend(self._build_sql_from_event_templates(templates))
 		
@@ -324,7 +408,6 @@ class ZeitgeistEngine:
 		Return modes:
 		 - 0: IDs.
 		 - 1: Events.
-		 - 2: (Timestamp, SubjectUri)
 		"""
 		t = time.time()
 		
@@ -338,8 +421,6 @@ class ZeitgeistEngine:
 			sql = "SELECT DISTINCT id FROM event_view"
 		elif return_mode == 1:
 			sql = "SELECT id FROM event_view"
-		elif return_mode == 2:
-			sql = "SELECT subj_uri, timestamp FROM event_view"
 		else:
 			raise NotImplementedError, "Unsupported return_mode."
 		
@@ -355,10 +436,13 @@ class ZeitgeistEngine:
 		
 		sql += (" ORDER BY timestamp DESC",
 			" ORDER BY timestamp ASC",
-			" GROUP BY subj_uri ORDER BY timestamp DESC",
-			" GROUP BY subj_uri ORDER BY timestamp ASC",
-			" GROUP BY subj_uri ORDER BY COUNT(subj_uri) DESC, timestamp DESC",
-			" GROUP BY subj_uri ORDER BY COUNT(subj_uri) ASC, timestamp ASC",
+			# thekorn: please note, event.subj_id == uri.id, as in
+			# the subj_id points directly to an entry in the uri table,
+			# so we are in fact grouping by subj_uris here
+			" GROUP BY subj_id ORDER BY timestamp DESC",
+			" GROUP BY subj_id ORDER BY timestamp ASC",
+			" GROUP BY subj_id ORDER BY COUNT(subj_id) DESC, timestamp DESC",
+			" GROUP BY subj_id ORDER BY COUNT(subj_id) ASC, timestamp ASC",
 			" GROUP BY actor ORDER BY COUNT(actor) DESC, timestamp DESC", 
 			" GROUP BY actor ORDER BY COUNT(actor) ASC, timestamp ASC",
 			" GROUP BY actor ORDER BY timestamp DESC",
@@ -367,7 +451,15 @@ class ZeitgeistEngine:
 			" GROUP BY subj_origin ORDER BY timestamp ASC",
 			" GROUP BY subj_origin ORDER BY COUNT(subj_origin) DESC, timestamp DESC",
 			" GROUP BY subj_origin ORDER BY COUNT(subj_origin) ASC, timestamp ASC",
-			" GROUP BY actor ORDER BY timestamp ASC")[order]
+			" GROUP BY actor ORDER BY timestamp ASC",
+			" GROUP BY subj_interpretation ORDER BY timestamp DESC",
+			" GROUP BY subj_interpretation ORDER BY timestamp ASC",
+			" GROUP BY subj_interpretation ORDER BY COUNT(subj_interpretation) DESC",
+			" GROUP BY subj_interpretation ORDER BY COUNT(subj_interpretation) ASC",
+			" GROUP BY subj_mimetype ORDER BY timestamp DESC",
+			" GROUP BY subj_mimetype ORDER BY timestamp ASC",
+			" GROUP BY subj_mimetype ORDER BY COUNT(subj_mimetype) DESC",
+			" GROUP BY subj_mimetype ORDER BY COUNT(subj_mimetype) ASC")[order]
 		
 		if max_events > 0:
 			sql += " LIMIT %d" % max_events
@@ -379,10 +471,7 @@ class ZeitgeistEngine:
 			result = [row[0] for row in result]
 		elif return_mode == 1:
 			log.debug("Found %d events IDs in %fs" % (len(result), time.time()- t))
-			result = self.get_events(ids=[row[0] for row in result], sender=sender)			
-		elif return_mode == 2:
-			log.debug("Found %d (uri,timestamp) tuples in %fs" % (len(result), time.time()- t))
-			result = map(lambda row: (row[0], row[1]), result)			
+			result = self.get_events(ids=[row[0] for row in result], sender=sender)	
 		else:
 			raise Exception("%d" % return_mode)
 		
@@ -394,12 +483,6 @@ class ZeitgeistEngine:
 	def find_events(self, *args):
 		return self._find_events(1, *args)
 	
-	def __add_window(self, _set, assoc, landmarks, windows):
-		if _set & landmarks: # intersection != 0
-			windows.append(_set)
-			for i in _set.difference(landmarks):
-				assoc[i] += 1
-	
 	def find_related_uris(self, timerange, event_templates, result_event_templates,
 		result_storage_state, num_results, result_type):
 		"""
@@ -411,81 +494,57 @@ class ZeitgeistEngine:
 		and `result_storage_state` are returned.
 		"""
 		
-		if result_type == 0 or result_type == 1:
-			
+		if result_type in (0, 1):
+			# Instead of using sliding windows we will be using a graph representation
+
 			t1 = time.time()
 			
-			if len(result_event_templates) == 0:
-				uris = self._find_events(2, timerange, result_event_templates,
-					result_storage_state, 0, ResultType.LeastRecentEvents)
-			else:
-				uris = self._find_events(2, timerange, result_event_templates + event_templates,
+			# We pick out the ids for relational event so we can set them as roots
+			# the ids are taken from the events that match the events_templates
+			ids = self._find_events(0, timerange, event_templates,
 					result_storage_state, 0, ResultType.LeastRecentEvents)
 			
-			assoc = defaultdict(int)
+			# Pick out the result_ids for the filtered results we would like to take into account
+			# the ids are taken from the events that match the result_event_templates
+			# if no result_event_templates are set we consider all results as allowed
+			result_ids = []
+			if len(result_event_templates) > 0:
+				result_ids = self._find_events(0, timerange, result_event_templates,
+						result_storage_state, 0, ResultType.LeastRecentEvents)
 			
-			landmarks = self._find_events(2, timerange, event_templates,
-					result_storage_state, 0, 4)
-			landmarks = set([unicode(event[0]) for event in landmarks])
+			# From here we create several graphs with the maximum depth of 2
+			# and push all the nodes and vertices (events) in one pot together
+			# FIXME: the depth should be adaptable 
+			pot = []
+			for id in ids:
+				for x in xrange(id-2,id+3):
+					if len(result_ids) == 0 or x in result_ids:
+						pot.append(x)
 			
-			latest_uris = dict(uris)
-			events = [unicode(u[0]) for u in uris]
-
-			furis = filter(lambda x: x[0] in landmarks, uris)
-			if len(furis) == 0:
-				return []
+			# Out of the pot we get all respected events and count which uris occur most
+			rows = self._cursor.execute("""
+				SELECT id, timestamp, subj_uri FROM event_view
+				WHERE id IN (%s)
+				""" % ",".join("%d" % id for id in pot)).fetchall()
 			
-			_min = min(furis, key=operator.itemgetter(1))
-			_max = max(furis, key=operator.itemgetter(1))
-			min_index = uris.index(_min) - WINDOW_SIZE
-			max_index = uris.index(_max) + WINDOW_SIZE
-			_min = _min[1]
-			_max = _max[1]
-			
-			if min_index < 0:
-				min_index = 0
-			if max_index > len(events):
-				max_index = -1
-				
-			func = self.__add_window
-			
-			if len(events) == 0 or len(landmarks) == 0:
-				return []
-			
-			windows = []
-	
-			if len(events) <= WINDOW_SIZE:
-				#TODO bug! windows is not defined, seems the algorithm never touches these loop
-				func(events, assoc, landmarks, windows)
-			else:
-				events = events[min_index:max_index]
-				offset = WINDOW_SIZE/2
-				
-				for i in xrange(offset):
-					func(set(events[0: offset - i]), assoc, landmarks, 
-						windows)
-					func(set(events[len(events) - offset + i: len(events)]),
-						assoc, landmarks, windows)
-					
-				it = iter(events)
-				result = tuple(islice(it, WINDOW_SIZE))
-				for elem in it:
-					result = result[1:] + (elem,)
-					func(set(result), assoc, landmarks, windows)
-					
-				
-			log.debug("FindRelatedUris: Finished sliding windows in %fs." % \
+			subject_uri_counter = defaultdict(int)
+			latest_uris = defaultdict(int)
+			for id, timestamp, uri in rows:
+				if id not in ids:
+					subject_uri_counter[uri] += 1
+					if latest_uris[uri] < timestamp:
+						latest_uris[uri] = timestamp
+							
+			log.debug("FindRelatedUris: Finished ranking subjects %fs." % \
 				(time.time()-t1))
 			
 			if result_type == 0:
-				sets = [[v, k] for k, v in assoc.iteritems()]
+				sets = subject_uri_counter.iteritems()
 			elif result_type == 1:
-				sets = [[latest_uris[k], k] for k in assoc]
+				sets = ((k, latest_uris[k]) for k in subject_uri_counter)
 				
-			sets.sort(reverse = True)
-			sets = map(lambda result: result[1], sets[:num_results])
-			
-			return sets
+			sets = sorted(sets, reverse=True, key=lambda x: x[1])[:num_results]
+			return map(lambda result: result[0], sets)
 		else:
 			raise NotImplementedError, "Unsupported ResultType."
 			
@@ -493,8 +552,9 @@ class ZeitgeistEngine:
 	def insert_events(self, events, sender=None):
 		t = time.time()
 		m = map(lambda e: self._insert_event_without_error(e, sender), events)
-		self._cursor.connection.commit()
-		log.debug("Inserted %d events in %fs" % (len(m), time.time()-t))
+		self._cursor.connection.commit() 
+		log.debug("Inserted %d events out of %d in %fs"\
+				 % (len([e for e in m if e>0]), len(m), time.time()-t))
 		return m
 	
 	def _insert_event_without_error(self, event, sender=None):
@@ -552,7 +612,12 @@ class ZeitgeistEngine:
 		try:
 			for subject in event.subjects:	
 				self._cursor.execute("""
-					INSERT INTO event VALUES (
+					INSERT INTO event (
+						id, timestamp, interpretation, manifestation, actor,
+						payload, subj_id,
+						subj_interpretation, subj_manifestation, subj_origin,
+						subj_mimetype, subj_text, subj_storage
+					) VALUES (
 						?, ?, ?, ?, ?, ?,
 						(SELECT id FROM uri WHERE value=?),
 						?, ?,
@@ -621,6 +686,11 @@ class ZeitgeistEngine:
 		""" % ",".join(str(int(_id)) for _id in ids))
 		timestamps = self._cursor.fetchone()
 
+		# Remove events from cache
+		for id in ids:
+			if id in self._event_cache:
+				del self._event_cache[id]
+		
 		# Make sure that we actually found some events with these ids...
 		# We can't do all(timestamps) here because the timestamps may be 0
 		if timestamps and timestamps[0] is not None and timestamps[1] is not None:
@@ -630,7 +700,6 @@ class ZeitgeistEngine:
 			log.debug("Deleted %s" % map(int, ids))
 			
 			self.extensions.apply_post_delete(ids, sender)
-			
 			return timestamps
 		else:
 			log.debug("Tried to delete non-existing event(s): %s" % map(int, ids))
