@@ -9,7 +9,7 @@
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
+# the Free Software Foundation, either version 2.1 of the License, or
 # (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
@@ -27,7 +27,6 @@ import os
 
 from _zeitgeist.engine import constants
 
-logging.basicConfig(level=logging.DEBUG)
 log = logging.getLogger("zeitgeist.sql")
 
 TABLE_MAP = {
@@ -48,9 +47,21 @@ class UnicodeCursor(sqlite3.Cursor):
 	
 	@staticmethod
 	def fix_unicode(obj):
+		if isinstance(obj, (int, long)):
+			# thekorn: as long as we are using the unary operator for timestamp
+			# related queries we have to make sure that integers are not
+			# converted to strings, same applies for long numbers.
+			return obj
 		if isinstance(obj, str):
 			obj = obj.decode("UTF-8")
-		return unicode(obj)		
+		# seif: Python’s default encoding is ASCII, so whenever a character with
+		# an ASCII value > 127 is in the input data, you’ll get a UnicodeDecodeError
+		# because that character can’t be handled by the ASCII encoding.
+		try:
+			obj = unicode(obj)
+		except UnicodeDecodeError, ex:
+			pass
+		return obj
 	
 	def execute(self, statement, parameters=()):
 		parameters = [self.fix_unicode(p) for p in parameters]
@@ -93,20 +104,20 @@ def _set_schema_version (cursor, schema_name, version):
 def _do_schema_upgrade (cursor, schema_name, old_version, new_version):
 	"""
 	Try and upgrade schema `schema_name` from version `old_version` to
-	`new_version`. This is done by checking for an upgrade module named
-	'_zeitgeist.engine.upgrades.$schema_name_$old_version_$new_version'
-	and executing the run(cursor) method of that module
+	`new_version`. This is done by executing a series of upgrade modules
+	named '_zeitgeist.engine.upgrades.$schema_name_$(i)_$(i+1)' and executing 
+	the run(cursor) method of those modules until new_version is reached
 	"""
-	# Fire of the right upgrade module
-	log.info("Upgrading database '%s' from version %s to %s. This may take a while" %
-	         (schema_name, old_version, new_version))
-	upgrader_name = "%s_%s_%s" % (schema_name, old_version, new_version)
-	module = __import__ ("_zeitgeist.engine.upgrades.%s" % upgrader_name)
-	eval("module.engine.upgrades.%s.run(cursor)" % upgrader_name)
-	
-	# Update the schema version
-	_set_schema_version(cursor, schema_name, new_version)
-	
+	for i in xrange(old_version, new_version):
+		# Fire of the right upgrade module
+		log.info("Upgrading database '%s' from version %s to %s. This may take a while" %
+		         (schema_name, i, i+1))
+		upgrader_name = "%s_%s_%s" % (schema_name, i, i+1)
+		module = __import__ ("_zeitgeist.engine.upgrades.%s" % upgrader_name)
+		eval("module.engine.upgrades.%s.run(cursor)" % upgrader_name)
+		
+		# Update the schema version
+		_set_schema_version(cursor, schema_name, i+1)
 	log.info("Upgrade succesful")
 
 def _check_core_schema_upgrade (cursor):
@@ -114,7 +125,7 @@ def _check_core_schema_upgrade (cursor):
 	# See if we have the right schema version, and try an upgrade if needed
 	core_schema_version = _get_schema_version(cursor, constants.CORE_SCHEMA)
 	if core_schema_version is not None:
-		if core_schema_version == constants.CORE_SCHEMA_VERSION:
+		if core_schema_version >= constants.CORE_SCHEMA_VERSION:
 			return True
 		else:
 			try:
@@ -138,17 +149,32 @@ def create_db(file_path):
 	"""Create the database and return a default cursor for it"""
 	start = time.time()
 	log.info("Using database: %s" % file_path)
+	new_database = not os.path.exists(file_path)
 	conn = sqlite3.connect(file_path)
 	conn.row_factory = sqlite3.Row
 	cursor = conn.cursor(UnicodeCursor)
+
+	# Seif: as result of the optimization story (LP: #639737) we are setting
+	# journal_mode to WAL if possible, this change is irreversible but
+	# gains us a big speedup, for more information see http://www.sqlite.org/wal.html
+	# FIXME: Set journal_mode to WAL when teamdecision has been take.
+	# cursor.execute("PRAGMA journal_mode = WAL")
+	cursor.execute("PRAGMA journal_mode = DELETE")
+	# Seif: another result of the performance tweaks discussed in (LP: #639737)
+	# we decided to set locking_mode to EXCLUSIVE, from now on only
+	# one connection to the database is allowed to revert this setting set locking_mode to NORMAL.
+	cursor.execute("PRAGMA locking_mode = EXCLUSIVE")
 	
 	# Always assume that temporary memory backed DBs have good schemas
 	if constants.DATABASE_FILE != ":memory:":
-		if _check_core_schema_upgrade (cursor):
+		if not new_database and _check_core_schema_upgrade(cursor):
 			_time = (time.time() - start)*1000
 			log.debug("Core schema is good. DB loaded in %sms" % _time)
 			return cursor
 	
+	# the following sql statements are only executed if a new database
+	# is created or an update of the core schema was done
+	log.debug("Updating sql schema")
 	# uri
 	cursor.execute("""
 		CREATE TABLE IF NOT EXISTS uri
@@ -355,10 +381,12 @@ def create_db(file_path):
 					AS payload,
 				(SELECT value FROM uri WHERE uri.id=event.subj_id)
 					AS subj_uri,
+				event.subj_id, -- #this directly points to an id in the uri table
 				event.subj_interpretation,
 				event.subj_manifestation,
+				event.subj_origin,
 				(SELECT value FROM uri WHERE uri.id=event.subj_origin)
-					AS subj_origin,
+					AS subj_origin_uri,
 				event.subj_mimetype,
 				(SELECT value FROM text WHERE text.id = event.subj_text)
 					AS subj_text,
@@ -433,6 +461,22 @@ class TableLookup(dict):
 		# Use this when fetching values which are supposed to be in the
 		# database already. Eg., in find_eventids.
 		return super(TableLookup, self).__getitem__(name)
+		
+def get_right_boundary(text):
+	""" returns the smallest string which is greater than `text` """
+	if not text:
+		# if the search prefix is empty we query for the whole range
+		# of 'utf-8 'unicode chars
+		return unichr(0x10ffff)
+	if isinstance(text, str):
+		# we need to make sure the text is decoded as 'utf-8' unicode
+		text = unicode(text, "UTF-8")
+	charpoint = ord(text[-1])
+	if charpoint == 0x10ffff:
+		# if the last character is the biggest possible char we need to
+		# look at the second last
+		return get_right_boundary(text[:-1])
+	return text[:-1] + unichr(charpoint+1)
 
 class WhereClause:
 	"""
@@ -453,6 +497,25 @@ class WhereClause:
 	OR = " OR "
 	NOT = "NOT "
 	
+	@staticmethod
+	def optimize_glob(column, table, prefix):
+		"""returns an optimized version of the GLOB statement as described
+		in http://www.sqlite.org/optoverview.html `4.0 The LIKE optimization`
+		"""
+		if isinstance(prefix, str):
+			# we need to make sure the text is decoded as 'utf-8' unicode
+			prefix = unicode(prefix, "UTF-8")
+		if not prefix:
+			# empty prefix means 'select all', no way to optimize this
+			sql = "SELECT %s FROM %s" %(column, table)
+			return sql, ()
+		elif all([i == unichr(0x10ffff) for i in prefix]):
+			sql = "SELECT %s FROM %s WHERE value >= ?" %(column, table)
+			return sql, (prefix,)
+		else:
+			sql = "SELECT %s FROM %s WHERE (value >= ? AND value < ?)" %(column, table)
+			return sql, (prefix, get_right_boundary(prefix))
+	
 	def __init__(self, relation, negation=False):
 		self._conditions = []
 		self.arguments = []
@@ -474,24 +537,18 @@ class WhereClause:
 			
 	def add_text_condition(self, column, value, like=False, negation=False, cache=None):
 		if like:
-			# thekorn: unfortunatly the data in event_view is a bit inconsistent
-			# e.g.:
-			# subj_uri and subj_origin are presented as string-values
-			# actor and subj_mimetype are ids
-			# (LP: #580601)
-			if column in ("subj_uri", "subj_origin"):
-				value_type = "value"
-			elif column in ("actor", "subj_mimetype"):
-				value_type = "id"
+			assert column in ("subj_uri", "subj_origin", "actor", "subj_mimetype"), \
+				"prefix search on the %r column is not supported by zeitgeist" %column
+			if column == "subj_uri":
+				# subj_id directly points to the id of an uri entry
+				view_column = "subj_id"
 			else:
-				raise AssertionError("We don't know how to handle this type of data")
-			# thekorn: this is a first (unoptimized version)
-			# see http://www.sqlite.org/optoverview.html '4.0 The LIKE optimization'
-			# for how this will look in the future
-			sql = "%s %sIN (SELECT %s FROM %s WHERE value GLOB ?)" \
-					%(column, self.NOT if negation else "", value_type, TABLE_MAP.get(column, column))
-			value += "*"
+				view_column = column
+			optimized_glob, value = self.optimize_glob("id", TABLE_MAP.get(column, column), value)
+			sql = "%s %sIN (%s)" %(view_column, self.NOT if negation else "", optimized_glob)
 		else:
+			if column == "subj_origin":
+				column = "subj_origin_uri"
 			sql = "%s %s= ?" %(column, "!" if negation else "")
 			if cache is not None:
 				value = cache[value]
