@@ -2,13 +2,15 @@
 
 # Zeitgeist
 #
-# Copyright © 2009-2010 Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
+# Copyright © 2009-2011 Siegfried-Angel Gevatter Pujals <siegfried@gevatter.com>
 # Copyright © 2009 Mikkel Kamstrup Erlandsen <mikkel.kamstrup@gmail.com>
-# Copyright © 2009 Markus Korn <thekorn@gmx.de>
+# Copyright © 2009-2011 Markus Korn <thekorn@gmx.de>
+# Copyright © 2011 Collabora Ltd.
+#             By Siegfried-Angel Gevatter Pujals <siegfried@gevatter.com>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
+# the Free Software Foundation, either version 2.1 of the License, or
 # (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
@@ -25,10 +27,9 @@ import dbus.mainloop.glib
 import logging
 import os.path
 import sys
-import logging
 import inspect
 
-from xml.dom.minidom import parseString as minidom_parse
+from xml.etree import ElementTree
 
 dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
 
@@ -37,46 +38,70 @@ from zeitgeist.datamodel import (Event, Subject, TimeRange, StorageState,
 
 SIG_EVENT = "asaasay"
 
-logging.basicConfig(level=logging.DEBUG)
 log = logging.getLogger("zeitgeist.client")
 
 class _DBusInterface(object):
 	"""Wrapper around dbus.Interface adding convenience methods."""
 
+	# We initialize those as sets in the constructor. Remember that we can't do
+	# that here because otherwise all instances would share their state.
+	_disconnect_callbacks = None
+	_reconnect_callbacks = None
+
 	@staticmethod
 	def get_members(introspection_xml):
 		"""Parses the XML context returned by Introspect() and returns
 		a tuple, where the first item is a list of all methods and the
 		second one a list of all signals for the related interface
 		"""
-		doc = minidom_parse(introspection_xml)
-		nodes = doc.getElementsByTagName("signal")
-		signals = [node.getAttribute("name") for node in nodes]
-		nodes = doc.getElementsByTagName("method")
-		methods = [node.getAttribute("name") for node in nodes]
+		xml = ElementTree.fromstring(introspection_xml)
+		nodes = xml.findall("interface/signal")
+		signals = [node.attrib["name"] for node in nodes]
+		nodes = xml.findall("interface/method")
+		methods = [node.attrib["name"] for node in nodes]
 		try:
 			methods.remove("Introspect") # Introspect is not part of the API
 		except ValueError:
 			pass
 		return methods, signals
 
-	def _disconnection_safe(self, meth):
+	def reconnect(self):
+		self.__proxy = dbus.SessionBus().get_object(self.__iface.requested_bus_name,
+			self.__object_path)
+		self.__iface = dbus.Interface(self.__proxy, self.__interface_name)
+
+	def _disconnection_safe(self, meth, *args, **kwargs):
 		"""
 		Executes the given method. If it fails because the D-Bus connection
 		was lost, attempts to recover it and executes the method again.
 		"""
-		try:
-			return meth()
-		except dbus.exceptions.DBusException, e:
+		
+		custom_error_handler = None
+		original_kwargs = dict(kwargs)
+
+		def reconnecting_error_handler(e):
 			error = e.get_dbus_name()
 			if error == "org.freedesktop.DBus.Error.ServiceUnknown":
-				self.__proxy = dbus.SessionBus().get_object(
-					self.__iface.requested_bus_name, self.__object_path)
-				self.__iface = dbus.Interface(self.__proxy,
-					self.__interface_name)
-				return meth()
+				self.reconnect()
+				# We don't use the reconnecting_error_handler here since that'd
+				# get us into an endless loop if Zeitgeist really isn't there.
+				return meth(*args, **original_kwargs)
 			else:
-				raise
+				if custom_error_handler is not None:
+					custom_error_handler(e)
+				else:
+					raise
+
+		if 'error_handler' in kwargs:
+			# If the method is being called asynchronously it'll call the given
+			# handler on failure instead of directly raising an exception.
+			custom_error_handler = kwargs['error_handler']
+			kwargs['error_handler'] = reconnecting_error_handler
+
+		try:
+			return meth(*args, **kwargs)
+		except dbus.exceptions.DBusException, e:
+			reconnecting_error_handler()
 
 	def __getattr__(self, name):
 		if name not in self.__methods:
@@ -86,8 +111,8 @@ class _DBusInterface(object):
 			Method wrapping around a D-Bus call, which attempts to recover
 			the connection to Zeitgeist if it got lost.
 			"""
-			return self._disconnection_safe(lambda:
-				getattr(self.__iface, name)(*args, **kwargs))
+			return self._disconnection_safe(getattr(self.__iface, name),
+				*args, **kwargs)
 		return _ProxyMethod
 
 	def connect(self, signal, callback, **kwargs):
@@ -101,16 +126,16 @@ class _DBusInterface(object):
 			**kwargs)
 
 	def connect_exit(self, callback):
-		"""Executes callback when the RemoteInterface exists"""
-		bus_obj = dbus.SessionBus().get_object(dbus.BUS_DAEMON_IFACE,
-			dbus.BUS_DAEMON_PATH)
-		bus_obj.connect_to_signal(
-			"NameOwnerChanged",
-			lambda *args: callback(),
-			dbus_interface=dbus.BUS_DAEMON_IFACE,
-			arg0=self.__iface.requested_bus_name, # only match what we want
-			arg2="", # only match services with no new owner
-		)
+		"""Executes callback when the remote interface disappears from the bus"""
+		self._disconnect_callbacks.add(callback)
+
+	def connect_join(self, callback):
+		"""
+		Executes callback when someone claims the Zeitgeist D-Bus name.
+		This may be used to perform some action if the daemon is restarted while
+		it was being used.
+		"""
+		self._reconnect_callbacks.add(callback)
 
 	@property
 	def proxy(self):
@@ -122,6 +147,21 @@ class _DBusInterface(object):
 		self.__object_path = object_path
 		self.__iface = dbus.Interface(proxy, interface_name)
 		self.__methods, self.__signals = self.get_members(proxy.Introspect())
+		
+		self._disconnect_callbacks = set()
+		self._reconnect_callbacks = set()
+		
+		# Listen to (dis)connection notifications, for connect_exit and connect_join
+		def name_owner_changed(connection_name):
+			if connection_name == "":
+				callbacks = self._disconnect_callbacks
+			else:
+				self.reconnect()
+				callbacks = self._reconnect_callbacks
+			for callback in callbacks:
+				callback()
+		dbus.SessionBus().watch_name_owner(self.__iface.requested_bus_name,
+			name_owner_changed)
 
 class ZeitgeistDBusInterface(object):
 	""" Central DBus interface to the Zeitgeist engine
@@ -143,10 +183,18 @@ class ZeitgeistDBusInterface(object):
 	def version(self):
 		"""Returns the API version"""
 		dbus_interface = self.__shared_state["dbus_interface"]
-		return dbus_interface._disconnection_safe(lambda:
+		return dbus_interface._disconnection_safe(
 			dbus_interface.proxy.get_dbus_method("Get",
-				dbus_interface=dbus.PROPERTIES_IFACE)(self.INTERFACE_NAME,
-					"version"))
+				dbus_interface=dbus.PROPERTIES_IFACE),
+			self.INTERFACE_NAME, "version")
+	
+	def extensions(self):
+		"""Returns active extensions"""
+		dbus_interface = self.__shared_state["dbus_interface"]
+		return dbus_interface._disconnection_safe(
+			dbus_interface.proxy.get_dbus_method("Get",
+				dbus_interface=dbus.PROPERTIES_IFACE),
+				self.INTERFACE_NAME, "extensions")
 	
 	def get_extension(cls, name, path, busname=None):
 		""" Returns an interface to the given extension.
@@ -293,6 +341,8 @@ class ZeitgeistClient:
 	DBus calls use the raw DBus API found in the ZeitgeistDBusInterface class.
 	"""
 	
+	_installed_monitors = []
+	
 	@staticmethod
 	def get_event_and_extra_arguments(arguments):
 		""" some methods of :class:`ZeitgeistClient` take a variable
@@ -311,6 +361,18 @@ class ZeitgeistClient:
 		self._iface = ZeitgeistDBusInterface()
 		self._registry = self._iface.get_extension("DataSourceRegistry",
 			"data_source_registry")
+		
+		# Reconnect all active monitors if the connection is reset.
+		def reconnect_monitors():
+			log.info("Reconnected to Zeitgeist engine...")
+			for monitor in self._installed_monitors:
+				self._iface.InstallMonitor(monitor.path,
+					monitor.time_range,
+					monitor.templates,
+					reply_handler=self._void_reply_handler,
+					error_handler=lambda err: log.warn(
+						"Error reinstalling monitor: %s" % err))
+		self._iface.connect_join(reconnect_monitors)
 	
 	def _safe_error_handler(self, error_handler, *args):
 		if error_handler is not None:
@@ -328,8 +390,15 @@ class ZeitgeistClient:
 				"Reply handler not callable, found %s" % reply_handler)
 		return self._void_reply_handler
 	
+	# Properties
+	
 	def get_version(self):
 		return [int(i) for i in self._iface.version()]
+		
+	def get_extensions(self):
+		return [unicode(i) for i in self._iface.extensions()]
+	
+	# Methods
 	
 	def insert_event (self, event, ids_reply_handler=None, error_handler=None):
 		"""
@@ -597,10 +666,9 @@ class ZeitgeistClient:
 		Alias for :meth:`find_events_for_templates`, for use when only
 		one template is needed.
 		"""
-		ev, arguments = self.get_event_and_extra_arguments(kwargs)
-		self.find_events_for_templates([ev],
+		self.find_events_for_templates([event_template],
 						events_reply_handler,
-						**arguments)
+						**kwargs)
 	
 	def find_events_for_values(self, events_reply_handler, **kwargs):
 		"""
@@ -807,6 +875,7 @@ class ZeitgeistClient:
 		                           reply_handler=self._void_reply_handler,
 		                           error_handler=lambda err: log.warn(
 									"Error installing monitor: %s" % err))
+		self._installed_monitors.append(mon)
 		return mon
 	
 	def remove_monitor (self, monitor, monitor_removed_handler=None):
@@ -845,8 +914,14 @@ class ZeitgeistClient:
 		self._iface.RemoveMonitor(path,
 		                          reply_handler=reply_handler,
 		                          error_handler=error_handler)
+		self._installed_monitors.remove(monitor)
 	
-	def register_data_source(self, unique_id, name, description, event_templates):
+	# Data-source related class variables
+	_data_sources = {}
+	_data_sources_callback_installed = False
+	
+	def register_data_source(self, unique_id, name, description,
+		event_templates, enabled_callback=None):
 		"""
 		Register a data-source as currently running. If the data-source was
 		already in the database, its metadata (name, description and
@@ -863,13 +938,58 @@ class ZeitgeistClient:
 		:param description: data-source description (may be translated)
 		:param event_templates: list of
 			:class:`Event <zeitgeist.datamodel.Event>` templates.
+		:param enabled_callback: method to call as response with the `enabled'
+			status of the data-source, and after that every time said status
+			is toggled. See set_data_source_enabled_callback() for more
+			information.
 		"""
-		# TODO: Make it possible to access the return value!
+		
+		self._data_sources[unique_id] = {'enabled': None, 'callback': None}
+		
+		if enabled_callback is not None:
+			self.set_data_source_enabled_callback(unique_id, enabled_callback)
+
+		def _data_source_enabled_cb(unique_id, enabled):
+			if unique_id not in self._data_sources:
+				return
+			self._data_sources[unique_id]['enabled'] = enabled
+			callback = self._data_sources[unique_id]['callback']
+			if callback is not None:
+				callback(enabled)
+		
+		def _data_source_register_cb(enabled):
+			_data_source_enabled_cb(unique_id, enabled)
+
+		if not self._data_sources_callback_installed:
+			self._registry.connect('DataSourceEnabled', _data_source_enabled_cb)
+			self._data_sources_callback_installed = True
+
 		self._registry.RegisterDataSource(unique_id, name, description,
 			event_templates,
-			reply_handler=self._void_reply_handler,
+			reply_handler=_data_source_register_cb,
 			error_handler=self._void_reply_handler) # Errors are ignored
 	
+	def set_data_source_enabled_callback(self, unique_id, enabled_callback):
+		"""
+		This method may only be used after having registered the given unique_id
+		with register_data_source before.
+		
+		It registers a method to be called whenever the `enabled' status of
+		the previously registered data-source changes.
+		
+		Remember that on some systems the DataSourceRegistry extension may be
+		disabled, in which case this method will have no effect.
+		"""
+		
+		if unique_id not in self._data_sources:
+			raise ValueError, 'set_data_source_enabled_callback() called before ' \
+			'register_data_source()'
+		
+		if not callable(enabled_callback):
+			raise TypeError, 'enabled_callback: expected a callable method'
+		
+		self._data_sources[unique_id]['callback'] = enabled_callback
+	
 	def _check_list_or_tuple(self, collection):
 		"""
 		Raise a ValueError unless 'collection' is a list or tuple
